/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Implementation of task thread pool.
 *
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <complib/cl_mpthreadpool.h>

static void cl_mp_thread_pool_run_task(void *context)
{
	static __thread boolean_t call_affinitiy_check = FALSE;
	cl_mp_thread_pool_t *p_mp_thread_pool = (cl_mp_thread_pool_t *)context;
	void *task = NULL;
	uint32_t num_of_threads = p_mp_thread_pool->thread_pool.running_count;

	pthread_mutex_lock(&p_mp_thread_pool->mutex);
	if (!call_affinitiy_check) {
		call_affinitiy_check = TRUE;
		cl_thread_validate_affinity(p_mp_thread_pool->active_cores,
					    num_of_threads,
					    p_mp_thread_pool->max_threads_per_core);
	}

	task = p_mp_thread_pool->task_queue[--p_mp_thread_pool->queue_len];
	pthread_mutex_unlock(&p_mp_thread_pool->mutex);

	p_mp_thread_pool->pfn_callback(task, p_mp_thread_pool->contex);

	pthread_mutex_lock(&p_mp_thread_pool->mutex);
	p_mp_thread_pool->num_pending_tasks--;
	if (!p_mp_thread_pool->num_pending_tasks)
		pthread_cond_signal(&p_mp_thread_pool->cond);

	pthread_mutex_unlock(&p_mp_thread_pool->mutex);
}

cl_status_t cl_mp_thread_pool_init(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
				   IN unsigned num_threads,
				   IN uint16_t max_threads_per_core,
				   IN const char *const name)
{
	cl_status_t status = CL_INSUFFICIENT_MEMORY;

	CL_ASSERT(p_mp_thread_pool);

	p_mp_thread_pool->max_threads_per_core = max_threads_per_core;
	pthread_mutex_init(&p_mp_thread_pool->mutex, NULL);
	pthread_cond_init(&p_mp_thread_pool->cond, NULL);

	status = cl_thread_pool_init(&p_mp_thread_pool->thread_pool, num_threads,
				     cl_mp_thread_pool_run_task, p_mp_thread_pool, name);
	if (status != CL_SUCCESS) {
		pthread_cond_destroy(&p_mp_thread_pool->cond);
		pthread_mutex_destroy(&p_mp_thread_pool->mutex);
		goto Exit;;
	}

	p_mp_thread_pool->active_cores = (uint16_t *) calloc(p_mp_thread_pool->thread_pool.running_count,
						  sizeof(*p_mp_thread_pool->active_cores));
	if (!p_mp_thread_pool->active_cores) {
		cl_mp_thread_pool_destroy(p_mp_thread_pool);
		goto Exit;
	}

	status = CL_SUCCESS;
Exit:
	return status;
}

void cl_mp_thread_pool_destroy(IN cl_mp_thread_pool_t * const p_mp_thread_pool)
{
	CL_ASSERT(p_mp_thread_pool);

	if (p_mp_thread_pool->active_cores)
		free(p_mp_thread_pool->active_cores);

	cl_thread_pool_destroy(&p_mp_thread_pool->thread_pool);
	pthread_cond_destroy(&p_mp_thread_pool->cond);
	pthread_mutex_destroy(&p_mp_thread_pool->mutex);
}

void cl_mp_thread_pool_array_apply(IN cl_mp_thread_pool_t * const p_mp_thread_pool,
				   IN void * tasks[],
				   IN unsigned num_tasks,
				   IN void (*pfn_callback) (void *, void *),
				   IN void *contex)
{
	int i;

	CL_ASSERT(p_mp_thread_pool);
	CL_ASSERT(pfn_callback);
	CL_ASSERT(tasks);

	pthread_mutex_lock(&p_mp_thread_pool->mutex);
	p_mp_thread_pool->pfn_callback = pfn_callback;
	p_mp_thread_pool->num_pending_tasks = num_tasks;
	p_mp_thread_pool->queue_len = num_tasks;
	p_mp_thread_pool->contex = contex;

	p_mp_thread_pool->task_queue = (void **)malloc(p_mp_thread_pool->queue_len * sizeof(void *));
	for (i = 0; i < num_tasks; i++) {
		p_mp_thread_pool->task_queue[i] = tasks[i];
		cl_thread_pool_signal(&p_mp_thread_pool->thread_pool);
	}

	pthread_cond_wait(&p_mp_thread_pool->cond, &p_mp_thread_pool->mutex);
	free(p_mp_thread_pool->task_queue);
	p_mp_thread_pool->task_queue = NULL;
	pthread_mutex_unlock(&p_mp_thread_pool->mutex);
}
